"""
A specification drawer is for visualizing combinatorial specification object
"""
import json
import os
import tempfile
import threading
import time
import uuid
import webbrowser
from copy import copy
from typing import TYPE_CHECKING, ClassVar, Dict, Iterable, List, Tuple

import requests
from logzero import logger
from typing_extensions import TypedDict

from .combinatorial_class import CombinatorialClass
from .strategies import EquivalencePathRule, Rule, VerificationRule

if TYPE_CHECKING:
    from .specification import CombinatorialSpecification

__all__ = ("SpecificationDrawer", "ForestSpecificationDrawer")


class TreantNode(TypedDict):
    innerHTML: str
    collapsable: bool
    collapsed: bool
    children: list


class TreantConfig(TypedDict):
    chart: dict
    nodeStructure: TreantNode
    tooltips: List[Dict[str, str]]


class SpecificationDrawer:
    """
    A specification drawer is for visualizing CombinatorialSpecification by making
    HTML file that contains an interactive tree using the Treant javascript library

    https://fperucic.github.io/treant-js/
    Treant library converts a json of TreantNodes to HTML code
    """

    def __init__(
        self,
        spec: "CombinatorialSpecification",
        levels_shown: int = 0,
        levels_expand: int = 0,
        verbose: bool = False,
    ):
        """
        Initialise SpecificationDrawer.

        OTHER INPUT:
            - 'levels_shown': number of levels displayed at the start.
            If 0 then the whole tree is displayed
            - 'levels_expand': number of levels displayed after expanding a node.
            If 0 then the rest of the tree is displayed
            - 'verbose': displays formal step below the node
            and json representation in tooltips
        """
        self.spec = spec
        self.tooltips: List[Dict[str, str]] = []
        self._rules_dict_copy = copy(spec.rules_dict)
        self.levels_shown = levels_shown
        self.levels_expand = levels_expand
        self.verbose = verbose
        self.tree = self._to_tree(spec.root)

    @staticmethod
    def _get_new_node_id() -> str:
        new_id = uuid.uuid1()
        return str(new_id)

    def _to_tree(
        self, comb_class: CombinatorialClass, rec_depth: int = 0
    ) -> TreantNode:
        """Create the subtree rooted in comb_class"""
        try:
            # checks if the rule has come before
            rule = self._rules_dict_copy.pop(comb_class)
            recursed = False
        except KeyError:
            # recursion
            rule = self.spec.get_rule(comb_class)
            recursed = True

        # if EqvPathRule then the node and its child are put into single node
        if isinstance(rule, EquivalencePathRule):
            child = rule.children[0]
            try:
                rule = self._rules_dict_copy.pop(child)
            except KeyError:
                # child rule has recurred before
                rule = self.spec.get_rule(child)
                recursed = True
            comb_classes = [comb_class, child]
        else:
            comb_classes = [comb_class]

        if recursed:
            children = []
        else:
            children = [self._to_tree(child, rec_depth + 1) for child in rule.children]

        # if a node has children we create a delimiter node between them
        # in order to separate tooltip describing the steps taking in the tree

        if children and isinstance(rule, Rule):
            delimiter_node, node_id = self._create_delimiter_node(
                rule, children, rec_depth
            )
            self._create_delimiter_tooltip(rule, node_id)
            children = [delimiter_node]

        html_node = self.comb_classes_to_html_node(comb_classes)
        treant_node, node_id = self._create_standard_node(html_node, children)
        self._create_standard_tooltip(comb_classes, node_id)
        return treant_node

    def _create_delimiter_tooltip(self, rule: Rule, node_identifier: str) -> None:
        """Creates hover over tooltip for delimiter node"""
        tooltip = {
            "content": f"<p>Formal step:<br/>{rule.formal_step}</p>",
            "selector": f"#node{node_identifier}",
        }
        self.tooltips.append(tooltip)

    def _create_standard_tooltip(
        self, comb_classes: List[CombinatorialClass], node_identifier: str
    ) -> None:
        """Creates hover over tooltip for standard node"""
        rule = self.spec.get_rule(comb_classes[0])
        if isinstance(rule, EquivalencePathRule):
            # need to use the rule to be able to see the whole eqv path
            rule_string = str(rule).replace("\n", "<br>")
            # get the rule that starts from the end of the equiv path
            rule = self.spec.get_rule(comb_classes[-1])
        else:
            rule_string = str(comb_classes[0]).replace("\n", "<br>")

        # labels
        labels = [str(self.spec.get_label(comb_class)) for comb_class in comb_classes]
        labels_string = ", ".join(labels)

        # comb_class json
        json_string = ""
        if self.verbose:
            comb_classes_json = [
                "<br>JSON:<br>" + json.dumps(comb_class.to_jsonable())
                for comb_class in comb_classes
            ]
            json_string = f'<p> {", ".join(comb_classes_json)} </p>'

        tooltip = {
            "content": f"""
                <p>Labels: {labels_string}</p>
                <pre>{rule_string}</pre>
                {json_string}""",
            "selector": f"#node{node_identifier}",
        }
        if isinstance(rule, VerificationRule):
            tooltip["content"] += f"<p>Verified: {rule.formal_step}</p>"
        self.tooltips.append(tooltip)

    def _create_standard_node(
        self, html_node: str, children: List[TreantNode]
    ) -> Tuple[TreantNode, str]:
        """Returns a tuple containing a standard treant node and its id."""
        new_id = self._get_new_node_id()
        treant_node = TreantNode(
            innerHTML=f'<div id="node{new_id}" data-toggle="tooltip">{html_node}</div>',
            collapsable=False,
            collapsed=False,
            children=children,
        )
        return treant_node, new_id

    def _create_delimiter_node(
        self,
        rule: Rule,
        children: List[TreantNode],
        rec_depth: int,
        additional_style: str = "",
    ) -> Tuple[TreantNode, str]:
        """
        Returns tuple containing delimiter node that describes the steps taken between
        the rules and the node id
        """
        new_id = self._get_new_node_id()
        symbol = rule.strategy.get_op_symbol()
        delimiter_html = (
            f'<div class="and-gate" id={new_id} '
            f'style="{additional_style}">{symbol}</div>'
        )

        # check if node should be collapsed or not

        # True if recursion depth is same level_shown
        is_on_levels_shown = rec_depth != 0 and rec_depth == self.levels_shown
        # True at every levels_expand after levels_shown
        is_on_levels_expand = (
            self.levels_expand != 0  # if 0 we don't collapse
            and rec_depth > self.levels_shown
            and (rec_depth + self.levels_expand) % self.levels_expand == 0
        )
        # collapses at levels_shown and at every levels_expand after that
        collapsed = is_on_levels_shown or is_on_levels_expand

        delimiter_node = TreantNode(
            innerHTML=f"""<div id="node{new_id}"
                data-toggle="tooltip">{delimiter_html}</div>""",
            collapsable=True,
            collapsed=collapsed,
            children=children,
        )
        return delimiter_node, new_id

    def comb_classes_to_html_node(
        self,
        comb_classes: List[CombinatorialClass],
        additional_style: str = "",
        additional_label_style: str = "",
    ) -> str:
        """Returns a representation of comb classes as a single html node string"""
        if not comb_classes:
            raise RuntimeError("comb_classes argument should not be empty")
        html = ""
        inner_style = ""
        # makes node green if instance of VerificationRule
        if isinstance(self.spec.get_rule(comb_classes[-1]), VerificationRule):
            additional_style += "border-color: Green; border-width: 3px;"
            additional_label_style += "background-color: #89d75d;"
        # Check if comb_class has html representation function
        try:
            nodes = [comb_class.to_html_representation() for comb_class in comb_classes]
        except NotImplementedError:
            nodes = [
                str(comb_class).replace("\n", "<br>") for comb_class in comb_classes
            ]
            if len(nodes) > 1:
                inner_style += "border: 1px solid;"

        for node_string in nodes:
            html += f"""<div class=inner-node-content style="{inner_style}">
                {node_string}</div>"""

        # add labels above the node
        labels = [str(self.spec.get_label(comb_class)) for comb_class in comb_classes]
        labels_string = ", ".join(labels)
        labels_html = f"""<div class=label
            style='{additional_style}{additional_label_style}'>{labels_string}</div>"""

        # add verbose below the node
        verbose_html = ""
        if self.verbose:
            formal_step = self.spec.get_rule(comb_classes[-1]).formal_step
            verbose_html = f"""<div class=verbose> {formal_step}</div>"""

        return f"""<div class=node-container style='{additional_style}'>{labels_html}
            <div class=node-content
            style='max-width:{300 * len(nodes)}px;{additional_style}'>{html}</div>
            {verbose_html}</div>"""

    def show(self) -> None:
        """Displays the CombinatorialSpecification tree in the web browser"""
        fsd = ForestSpecificationDrawer([self])
        fsd.show()

    def share(self) -> None:
        """
        Upload the html of the specification on a file server and displays a link to the
        file.
        """
        fsd = ForestSpecificationDrawer([self])
        fsd.share()

    def to_html(self) -> str:
        """Returns a html string that contains the whole tree"""
        fsd = ForestSpecificationDrawer([self])
        return fsd.to_html()

    def export_html(self, file_name: str) -> None:
        """Creates a html file in current directory"""
        fsd = ForestSpecificationDrawer([self])
        fsd.export_html(file_name)

    def to_treant_json(self) -> str:
        """returns a json input for Treant library"""
        fsd = ForestSpecificationDrawer([self])
        return json.dumps(fsd.add_tree_config(0))


class ForestSpecificationDrawer:
    """
    ForestSpecificationDrawer is for drawing multiple SpecificationDrawer in
    one single HTML file by navigating them with tabs
    """

    def __init__(self, sd_list: Iterable[SpecificationDrawer]):
        self.sd_list = list(sd_list)

    def add_tree_config(self, sd_index: int) -> TreantConfig:
        """Returns the tree and tooltip as TreantConfig for treant library"""
        return TreantConfig(
            nodeStructure=self.sd_list[sd_index].tree,
            tooltips=self.sd_list[sd_index].tooltips,
            chart={
                "maxDepth": 10000,
                "container": f"#spec-tree{sd_index}",
                "connectors": {"type": "bCurve", "style": {}},
                "nodeAlign": "BOTTOM",
                "levelSeparation": 35,
                "siblingSeparation": 30,
                "connectorsSpeed ": 10,
                "animation": {"nodeSpeed": 400, "connectorsSpeed": 200},
                "callback": {
                    # can't send functions over so it
                    # is dealt with on javascript side
                    "onTreeLoaded": "",
                },
            },
        )

    @staticmethod
    def _read_file(filename: str) -> str:
        with open(filename, "r", encoding="UTF8") as f:
            return f.read()

    def _get_static_files(self) -> str:
        """returns stylesheet and js script as a single string"""
        # Path to static files
        psf = os.path.join(os.path.dirname(__file__), "resources/static/")

        treant_stylesheet = self._read_file(os.path.join(psf, "css/treant.css"))
        spec_stylesheet = self._read_file(os.path.join(psf, "css/specification.css"))
        bootstrap_stylesheet = self._read_file(os.path.join(psf, "css/bootstrap.css"))

        jquery_min_script = self._read_file(os.path.join(psf, "js/jquery.min.js"))
        bs_min_script = self._read_file(os.path.join(psf, "js/bootstrap.bundle.min.js"))
        treant_script = self._read_file(os.path.join(psf, "js/Treant.js"))
        raphael_script = self._read_file(os.path.join(psf, "js/raphael.js"))

        static_files = f"""<style>
        {bootstrap_stylesheet}
        {treant_stylesheet}
        {spec_stylesheet}
        </style>
        <script>{jquery_min_script}</script>
        <script>{bs_min_script}</script>
        <script>{treant_script}</script>
        <script>{raphael_script}</script>"""
        return static_files

    def _get_tree_content(self) -> Tuple[str, str, str]:
        """return a tuble of html string containing tabs, tab_content, json_input"""
        tabs = ""
        tab_content = ""
        forest_input = []
        for i in range(len(self.sd_list)):
            tabs += f"""
            <li class='nav-item'>
                <a data-toggle="tab" id='spectab{i}' href='#spectabcontent{i}'
                    class="nav-link {"active" if i == 0 else ""}"
                    contenteditable="true">
                Spec {i}
                </a>
            </li>
            """
            tab_content += f"""
            <div id=spectabcontent{i}
                class='tab-pane {"active show" if i == 0 else ""}'>
                <div id="spec-tree{i}" class="Treant Treant-loaded"></div>
            </div>
            """
            forest_input.append(self.add_tree_config(i))
        json_input = json.dumps(forest_input)
        return (tabs, tab_content, json_input)

    def to_html(self) -> str:
        """Returns a html string that contains the whole forest"""
        static_files = self._get_static_files()
        tabs, tab_content, json_input = self._get_tree_content()

        html_string = f"""
        <!DOCTYPE html><html><head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
        <title>Specification</title>
        {static_files}
        <div>
            <ul class="nav nav-tabs">
                {tabs}
            </ul>

            <div class='tab-content mt-2'>
                {tab_content}
            </div>
        </div>
        <script>
        var tooltip_opened = ""
        function setup_tooltips_event(tooltips) {{
            for (let k = 0; k < tooltips.tooltips.length; k++) {{
                $(tooltips.tooltips[k].selector).tooltip({{
                    template: `<div class="tooltip" role="tooltip">
                        <div class="tooltip-arrow"></div>
                        <div class="tooltip-inner"></div>
                    </div>`,
                    title: tooltips.tooltips[k].content,
                    html: true,
                    placement: "right",
                    fallbackPlacement: "flip",
                    animation: false,
                    trigger: "click",
                    viewport: {{
                        selector: 'body',
                    }}
                }});
                $(tooltips.tooltips[k].selector).on('show.bs.tooltip', function() {{
                    $(tooltip_opened).tooltip('hide')
                    tooltip_opened = tooltips.tooltips[k].selector
                }})
            }}
        }}
        let json_input = {json_input}

        let has_loaded = new Array(json_input.length);
        for (let i = 0; i < json_input.length; i++) {{
            let chart_config = json_input[i];
            chart_config.chart.callback.onTreeLoaded = function () {{}};
            let tree = new Treant(chart_config);
            if (i == 0) {{
                setup_tooltips_event(chart_config);
            }}
            $('#spectab' + i).on('shown.bs.tab', function (e) {{
                if (!has_loaded[i]) {{
                    tree.tree.reload();
                    setup_tooltips_event(chart_config);
                }}
                has_loaded[i] = true
            }});
            has_loaded[0] = true
        }}</script></body></html>"""
        return html_string

    @staticmethod
    def export_html(html: str, file_name: str = "tree.html") -> None:
        """Creates a html file in current directory"""
        text_file = open(file_name, "w", encoding="UTF8")
        text_file.write(html)
        text_file.close()

    def show(self) -> None:
        """Displays the forest in the web browser"""
        html_string = self.to_html()
        viewer = HTMLViewer()
        viewer.open_html(html_string)

    def share(self) -> None:
        """
        Upload the html of the forest on a file server and displays a link to the
        file.
        """
        html_string = self.to_html()

        with tempfile.NamedTemporaryFile(
            "r+", suffix=".html", delete=False, encoding="UTF8"
        ) as html_file:
            html_file.write(html_string)
            file_path = html_file.name

        # ask gofile.io which server it wants us to use
        logger.info("Sending specification to file host.")
        req1 = requests.get("https://apiv2.gofile.io/getServer")
        server = req1.json()["data"]["server"]
        upload_url = f"https://{server}.gofile.io/upload"

        with open(file_path, "rb") as f:
            req2 = requests.post(upload_url, files={"filesUploaded": f})

        os.remove(file_path)
        file_code = req2.json()["data"]["code"]
        _, file_name = os.path.split(file_path)

        file_enable_url = f"https://gofile.io/d/{file_code}"
        file_url = f"https://{server}.gofile.io/download/{file_code}/{file_name}"

        print(f"First click this: {file_enable_url}")
        print(f"Then this link will work: {file_url}")


class HTMLViewer:
    """A class for opening html text in browser."""

    _THREAD_WAIT_TIME: ClassVar[float] = 4  # seconds

    @staticmethod
    def _remove_file_thread(fname: str) -> None:
        time.sleep(HTMLViewer._THREAD_WAIT_TIME)
        if os.path.exists(fname):
            os.remove(fname)
            logger.info("specification html file removed")

    @staticmethod
    def _remove_file(fname: str) -> None:
        threading.Thread(target=HTMLViewer._remove_file_thread, args=(fname,)).start()

    @staticmethod
    def open_html(html: str) -> None:
        """Open and render html string in browser."""
        with tempfile.NamedTemporaryFile(
            "r+", suffix=".html", delete=False, encoding="UTF8"
        ) as html_file:
            html_file.write(html)
            webbrowser.open_new_tab(f"file://{html_file.name}")
            logger.info("Opening specification in browser")
            HTMLViewer._remove_file(html_file.name)

    @staticmethod
    def open_svg(svg: str) -> None:
        """Open and render svg image string in browser."""
        HTMLViewer.open_html(f"<html><body>{svg}</body></html>")



from .constructor import CartesianProduct, Constructor, DisjointUnion
from .rule import EquivalencePathRule, ReverseRule, Rule, VerificationRule
from .strategy import (
    AbstractStrategy,
    AtomStrategy,
    CartesianProductStrategy,
    DisjointUnionStrategy,
    EmptyStrategy,
    Strategy,
    StrategyFactory,
    SymmetryStrategy,
    VerificationStrategy,
)
from .strategy_pack import StrategyPack

__all__ = [
    "CartesianProduct",
    "Constructor",
    "DisjointUnion",
    "EquivalencePathRule",
    "ReverseRule",
    "Rule",
    "VerificationRule",
    "AbstractStrategy",
    "AtomStrategy",
    "CartesianProductStrategy",
    "DisjointUnionStrategy",
    "EmptyStrategy",
    "Strategy",
    "StrategyFactory",
    "SymmetryStrategy",
    "VerificationStrategy",
    "StrategyPack",
]



"""
A queue of labels.
"""
import abc
from collections import Counter, deque
from typing import Counter as CounterType
from typing import Deque, Iterator, List, Set, Tuple

import tabulate

from comb_spec_searcher.exception import NoMoreClassesToExpandError
from comb_spec_searcher.strategies.strategy_pack import StrategyPack
from comb_spec_searcher.typing import WorkPacket


class CSSQueue(abc.ABC):
    """
    A queue of labels.
    """

    def __init__(self, pack: StrategyPack):
        self.inferral_strategies = tuple(pack.inferral_strats)
        self.initial_strategies = tuple(pack.initial_strats)
        self.expansion_strats = tuple(tuple(x) for x in pack.expansion_strats)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CSSQueue):
            return NotImplemented
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    @abc.abstractmethod
    def add(self, label: int) -> None:
        """Add a label to the queue."""

    @abc.abstractmethod
    def set_not_inferrable(self, label: int) -> None:
        """You should avoid yielding label with an inferral strategy in future."""

    @abc.abstractmethod
    def set_verified(self, label: int) -> None:
        """Label was verified, how should the queue react?"""

    @abc.abstractmethod
    def set_stop_yielding(self, label: int) -> None:
        """You should stop yielding label. CSS indicated that this label should
        no longer be yielded, e.g., expanding another symmetry or expanding
        only children"""

    @abc.abstractmethod
    def do_level(self) -> Iterator[WorkPacket]:
        """Return a 'level'. The definition of level can be chosen by the user,
        but it determines the functionality of the do_level method in CSS."""

    @abc.abstractmethod
    def status(self) -> str:
        """Return a string that indicates that current status of the queue."""

    def __iter__(self) -> Iterator[WorkPacket]:
        return self

    @abc.abstractmethod
    def __next__(self) -> WorkPacket:
        """
        Yield the combinatorial classes in queue.
        It should yield triples (label, strategies, inferral)
        - label is the next thing to be expanded
        - strategies is the list of strategies to expand label with
        - inferral boolean tells the CSS if it is a inferral strategies. With
        inferral set to True, CSS will only apply the next in the list to the
        children, and assume you are ignore the parent from now on. The
        strategies will be applied cyclically until no change.
        """
        raise NotImplementedError


class DefaultQueue(CSSQueue):
    """
    The default queue used by CSS.
    """

    def __init__(self, pack: StrategyPack):
        super().__init__(pack)
        self.working: Deque[int] = deque()
        self.next_level: CounterType[int] = Counter()
        self.curr_level: Tuple[Deque[int], ...] = tuple(
            deque() for _ in self.expansion_strats
        )
        # One extra deque to be able to set ignore
        self.curr_level = self.curr_level + (deque(),)
        self._inferral_expanded: Set[int] = set()
        self._initial_expanded: Set[int] = set()
        self.ignore: Set[int] = set()
        self.queue_sizes: List[int] = []
        self.staging: Deque[WorkPacket] = deque([])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DefaultQueue):
            return NotImplemented
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    @property
    def levels_completed(self):
        """Return the number of times swapped from curr to next."""
        return len(self.queue_sizes)

    def add(self, label: int) -> None:
        if self.can_do_inferral(label) or self.can_do_initial(label):
            self.working.append(label)
        elif label not in self.ignore:
            self.next_level.update((label,))

    def set_verified(self, label) -> None:
        self.set_stop_yielding(label)

    def set_not_inferrable(self, label: int) -> None:
        """Mark the label such that it's not expanded with inferral anymore"""
        if label not in self.ignore:
            self._inferral_expanded.add(label)

    def set_not_initial(self, label: int) -> None:
        """Mark the label such that it's not expanded with initial anymore"""
        if label not in self.ignore:
            self._initial_expanded.add(label)

    def set_stop_yielding(self, label: int) -> None:
        self.ignore.add(label)
        # can remove it elsewhere to keep sets "small"
        self._inferral_expanded.discard(label)
        self._initial_expanded.discard(label)
        self.next_level.pop(label, None)

    def can_do_inferral(self, label: int) -> bool:
        """Return true if inferral strategies can be applied."""
        return bool(self.inferral_strategies) and label not in self._inferral_expanded

    def can_do_initial(self, label: int) -> bool:
        """Return true if initial strategies can be applied."""
        return bool(self.initial_strategies) and label not in self._initial_expanded

    def can_do_expansion(self, label: int, idx: int) -> bool:
        """Return true if expansion strategies can be applied."""
        return label not in self.ignore and idx < len(self.expansion_strats)

    def _populate_staging(self) -> None:
        """
        Populate the staging queue that is used by next to return WorkPacket.
        """
        while not self.staging and self.working:
            self.staging.extend(self._iter_helper_working())
        while not self.staging:
            if not any(self.curr_level):
                self._change_level()
            self.staging.extend(self._iter_helper_curr())

    def _change_level(self) -> None:
        assert not self.staging, "Can't change level is staging is not empty"
        assert not self.working, "Can't change level is working is not empty"
        assert not any(self.curr_level), "Can't change level is curr_level is not empty"
        self.curr_level[0].extend(
            label for label, _ in sorted(self.next_level.items(), key=lambda x: -x[1])
        )
        if not any(self.curr_level):
            raise StopIteration
        self.queue_sizes.append(len(self.curr_level[0]))
        self.next_level = Counter()

    def _iter_helper_curr(self) -> Iterator[WorkPacket]:
        assert any(self.curr_level), "The current queue is empty"
        # pylint: disable=stop-iteration-return
        idx, label = next(
            (
                (idx, queue.popleft())
                for idx, queue in enumerate(self.curr_level)
                if queue
            )
        )
        if idx == len(self.expansion_strats):
            self.set_stop_yielding(label)
            return
        for strat in self.expansion_strats[idx]:
            yield WorkPacket(label, (strat,), False)
        self.curr_level[idx + 1].append(label)

    def _iter_helper_working(self) -> Iterator[WorkPacket]:
        label = self.working.popleft()
        if self.can_do_inferral(label):
            yield WorkPacket(label, self.inferral_strategies, True)
            self.set_not_inferrable(label)
        if self.can_do_initial(label):
            for strat in self.initial_strategies:
                yield WorkPacket(label, (strat,), False)
            self.set_not_initial(label)
        self.next_level.update((label,))

    def __next__(self) -> WorkPacket:
        while True:
            while self.staging:
                wp = self.staging.popleft()
                if wp.label not in self.ignore:
                    return wp
            self._populate_staging()

    def do_level(self) -> Iterator[WorkPacket]:
        """
        An iterator of all combinatorial classes in the current queue.

        Will swap next queue to current after iteration.
        """
        curr_level = self.levels_completed
        while curr_level == self.levels_completed:
            try:
                yield next(self)
            except StopIteration as e:
                if curr_level == self.levels_completed:
                    raise NoMoreClassesToExpandError from e
                return

    def status(self) -> str:
        status = f"Queue status (currently on level {self.levels_completed}):\n"
        table: List[Tuple[str, str]] = []
        table.append(("working", f"{len(self.working):,d}"))
        for idx, queue in enumerate(self.curr_level[:-1]):
            table.append((f"current (set {idx+1})", f"{len(queue):,d}"))
        table.append(("next", f"{len(self.next_level):,d}"))
        status += "    "
        headers = ("Queue", "Size")
        colalign = ("left", "right")
        status += (
            tabulate.tabulate(table, headers=headers, colalign=colalign).replace(
                "\n", "\n    "
            )
            + "\n"
        )
        status += "\tThe size of the current queues at each level: {}".format(
            ", ".join(str(i) for i in self.queue_sizes)
        )
        return status



"""
A database for rules.
"""
import abc
from collections import defaultdict
from typing import Dict, Iterable, Iterator, List, MutableMapping, Optional, Set, Tuple

import tabulate
from logzero import logger

from comb_spec_searcher.equiv_db import EquivalenceDB
from comb_spec_searcher.exception import SpecificationNotFound
from comb_spec_searcher.strategies import AbstractStrategy, VerificationRule
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.tree_searcher import (
    Node,
    iterative_proof_tree_finder,
    iterative_prune,
    proof_tree_generator_dfs,
    prune,
    random_proof_tree,
    smallish_random_proof_tree,
)
from comb_spec_searcher.typing import RuleKey, SpecificationLabelsAndStrats

__all__ = ["RuleDBBase", "RuleDB"]


class RuleDBBase(abc.ABC):
    """A database for rules found."""

    def __init__(self) -> None:
        self.equivdb = EquivalenceDB()

    @property
    @abc.abstractmethod
    def rule_to_strategy(self) -> MutableMapping[RuleKey, AbstractStrategy]:
        pass

    def __eq__(self, other: object) -> bool:
        """Check if all stored information is the same."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return bool(self.rule_to_strategy == other.rule_to_strategy)

    def add(self, start: int, ends: Iterable[int], rule: AbstractRule) -> None:
        """
        Add a rule to the database.

        - start is a single integer.
        - ends is a tuple of integers, representing the non-empty children.
        - rule is a Rule that creates start -> ends.
        """
        ends = tuple(sorted(ends))
        if isinstance(rule, VerificationRule):
            self.set_verified(start)
        is_equiv = len(ends) == 1 and rule.strategy.can_be_equivalent()
        if is_equiv:
            self.set_equivalent(start, ends[0])
        if len(ends) != 1 or is_equiv or not self.are_equivalent(start, ends[0]):
            # to avoid overwriting an equivalence rule with a non-equivalence
            # rule, we only save if an equivalence rule, or does not have the
            # same start -> ends as some equivalence rule.
            self.rule_to_strategy[(start, ends)] = rule.strategy

    def is_verified(self, label: int) -> bool:
        """Return True if label has been verified."""
        return bool(self.equivdb.is_verified(label))

    def set_verified(self, label: int) -> None:
        """Mark label as verified."""
        self.equivdb.set_verified(label)

    def are_equivalent(self, label: int, other: int) -> bool:
        """Return true if label and other are equivalent."""
        return bool(self.equivdb.equivalent(label, other))

    def set_equivalent(self, label: int, other: int) -> None:
        """Mark label and other as equivalent."""
        self.equivdb.union(label, other)

    def rules_up_to_equivalence(self) -> Dict[int, Set[Tuple[int, ...]]]:
        """Return a defaultdict containing all rules up to the equivalence."""
        rules_dict: Dict[int, Set[Tuple[int, ...]]] = defaultdict(set)
        for start, ends in self:
            rules_dict[self.equivdb[start]].add(
                tuple(sorted(self.equivdb[e] for e in ends))
            )
        return rules_dict

    def all_rules(self) -> Iterator[Tuple[int, Tuple[int, ...], AbstractStrategy]]:
        """Yield all the rules found so far."""
        for start, ends in self:
            yield start, ends, self.rule_to_strategy[(start, ends)]

    def __iter__(self) -> Iterator[Tuple[int, Tuple[int, ...]]]:
        """Iterate through rules as the pairs (start, end)."""
        for start, ends in self.rule_to_strategy:
            if len(ends) != 1 or not self.are_equivalent(start, ends[0]):
                yield start, ends

    def contains(self, start: int, ends: Tuple[int, ...]) -> bool:
        """Return true if the rule start -> ends is in the database."""
        ends = tuple(sorted(ends))
        return (start, ends) in self.rule_to_strategy

    def status(self, elaborate: bool) -> str:
        """Return a string describing the status of the rule database."""
        status = "RuleDB status:\n"
        table: List[Tuple[str, str]] = []
        table.append(("Combinatorial rules", f"{len(self.rule_to_strategy):,d}"))
        if elaborate:
            eqv_labels = set()
            strategy_verified_labels = set()
            verified_labels = set()
            eqv_rules = set()
            for start, ends in self.rule_to_strategy:
                if not ends:
                    strategy_verified_labels.add(start)
                if self.is_verified(start):
                    verified_labels.add(start)
                eqv_start = self.equivdb[start]
                eqv_ends = tuple(sorted(self.equivdb[label] for label in ends))
                eqv_labels.add(eqv_start)
                eqv_labels.update(eqv_ends)
                eqv_rules.add((eqv_start, eqv_ends))
            table.append(
                ("Combintorial rules up to equivalence", f"{len(eqv_rules):,d}")
            )
            table.append(
                (
                    "Strategy verified combinatorial classes",
                    f"{len(strategy_verified_labels):,d}",
                )
            )
            table.append(
                (
                    "Verified combinatorial classes",
                    f"{len(verified_labels):,d}",
                )
            )
            table.append(
                (
                    "combinatorial classes up to equivalence",
                    f"{len(eqv_labels):,d}",
                )
            )
        status += "    "
        status += tabulate.tabulate(
            table, headers=("", "Total number"), colalign=("left", "right")
        ).replace("\n", "\n    ")
        return status

    ################################################################
    # Below are methods for finding a combinatorial specification. #
    ################################################################

    def has_specification(self, label: int) -> bool:
        """Return True if a specification has been found, false otherwise."""
        return self.is_verified(label)

    def rule_from_equivalence_rule(
        self, eqv_start: int, eqv_ends: Iterable[int]
    ) -> Optional[Tuple[int, Tuple[int, ...]]]:
        """
        Return a rule that satisfies the equivalence rule.

        Returns None if no such rule exists.
        """
        eqv_start = self.equivdb[eqv_start]
        eqv_ends = tuple(sorted(map(self.equivdb.__getitem__, eqv_ends)))
        for rule in self:
            start, ends = rule
            temp_start = self.equivdb[start]
            temp_ends = tuple(sorted(map(self.equivdb.__getitem__, ends)))
            if eqv_start == temp_start and eqv_ends == temp_ends:
                return start, ends
        return None

    def rule_from_equivalence_rule_dict(
        self, eqv_rules: List[RuleKey]
    ) -> Dict[RuleKey, RuleKey]:
        """
        Return a dictionary pointing from an equivalence rule to an actual rule.
        """
        res: Dict[RuleKey, RuleKey] = {}
        for start, ends in self.rule_to_strategy:
            eqv_start = self.equivdb[start]
            eqv_ends = tuple(sorted(map(self.equivdb.__getitem__, ends)))
            eqv_key = (eqv_start, eqv_ends)
            if eqv_key in eqv_rules:
                res[eqv_key] = (start, ends)
        return res

    def find_specification(
        self, label: int, minimization_time_limit: float, iterative: bool = False
    ) -> Node:
        """Search for a specification based on current data found."""
        rules_dict = self.rules_up_to_equivalence()
        # Prune all unverified labels (recursively)
        if iterative:
            rules_dict = iterative_prune(rules_dict, root=label)
        else:
            prune(rules_dict)  # this function removes rules not in a specification.

        # only verified labels in rules_dict, in particular, there is a
        # specification if a label is in the rules_dict
        for ver_label in rules_dict.keys():
            self.set_verified(ver_label)

        if self.equivdb[label] in rules_dict:
            logger.info("Specification detected.")
            if iterative:
                specification = iterative_proof_tree_finder(
                    rules_dict, root=self.equivdb[label]
                )
            else:
                logger.info(
                    "Minimizing for %s seconds.", round(minimization_time_limit)
                )
                specification = smallish_random_proof_tree(
                    rules_dict, self.equivdb[label], minimization_time_limit
                )
            logger.info(
                "Found specification with %s rules.", len(specification.labels())
            )
        else:
            raise SpecificationNotFound("No specification for label {}".format(label))
        return specification

    def get_specification_rules(
        self, label: int, minimization_time_limit: float, iterative: bool = False
    ) -> SpecificationLabelsAndStrats:
        """
        Return a list of pairs (label, rule) which form a specification.
        The specification returned is random, so two calls to the function
        may result in a a different output.
        """
        proof_tree_node = self.find_specification(
            label=label,
            iterative=iterative,
            minimization_time_limit=minimization_time_limit,
        )
        return self._get_specification_rules(label, proof_tree_node)

    def _get_specification_rules(
        self, label: int, proof_tree_node: Node
    ) -> SpecificationLabelsAndStrats:
        children: Dict[int, Tuple[int, ...]] = dict()
        internal_nodes = set([label])
        logger.info("Computing rule <-> equivalence rule mapping.")
        eqv_rules = [
            (node.label, tuple(sorted(child.label for child in node.children)))
            for node in proof_tree_node.nodes()
        ]
        rule_from_equivalence_rules = self.rule_from_equivalence_rule_dict(eqv_rules)
        logger.info("Finding actual rules.")
        for eqv_start, eqv_ends in eqv_rules:
            rule = rule_from_equivalence_rules.get((eqv_start, eqv_ends))
            if rule is not None:
                start, ends = rule
                children[start] = ends
                internal_nodes.update(ends)
        res = []
        eqv_paths = []
        logger.info("Computing strategies used to create rules.")
        for start, ends in children.items():
            for eqv_label in internal_nodes:
                if self.are_equivalent(start, eqv_label):
                    path = self.equivdb.find_path(eqv_label, start)
                    for a, b in zip(path[:-1], path[1:]):
                        try:
                            strategy = self.rule_to_strategy[(a, (b,))]
                            res.append((a, strategy))
                        except KeyError:
                            strategy = self.rule_to_strategy[(b, (a,))]
                            res.append((b, strategy))
                    if len(path) > 1:
                        eqv_paths.append(path)
            strategy = self.rule_to_strategy[(start, ends)]
            res.append((start, strategy))
        return res, eqv_paths

    def all_specifications(
        self, label: int, iterative: bool = False
    ) -> Iterator[SpecificationLabelsAndStrats]:
        """
        A generator that yields all specifications in the universe for
        the given label.
        """
        if iterative:
            raise NotImplementedError(
                "There is no method for yielding all iterative proof trees."
            )
        rules_dict = self.rules_up_to_equivalence()
        # Prune all unverified labels (recursively)
        prune(rules_dict)

        if self.equivdb[label] in rules_dict:
            proof_trees = proof_tree_generator_dfs(rules_dict, root=self.equivdb[label])
        for proof_tree_node in proof_trees:
            yield self._get_specification_rules(label, proof_tree_node)

    def get_smallest_specification(
        self, label: int, iterative: bool = False
    ) -> SpecificationLabelsAndStrats:
        """
        Return the smallest specification in the universe for label. It uses
        exponential search to find it.
        This doesn't consider the length of the equivalence paths.
        """
        if iterative:
            raise NotImplementedError(
                "There is no method for finding smallest iterative proof trees."
            )
        rules_dict = self.rules_up_to_equivalence()
        prune(rules_dict)

        if not self.equivdb[label] in rules_dict:
            raise SpecificationNotFound("No specification for label {}".format(label))
        tree = random_proof_tree(rules_dict, root=self.equivdb[label])
        minimum = 1
        maximum = len(tree)
        logger.info(
            "Found a specification of size %s. Looking for the smallest.", len(tree)
        )
        # Binary search to find a smallest proof tree.
        while minimum < maximum:
            middle = (minimum + maximum) // 2
            logger.info("Looking for specification of size %s", middle)
            try:
                tree = next(
                    proof_tree_generator_dfs(
                        rules_dict, root=self.equivdb[label], maximum=middle
                    )
                )
                maximum = min(middle, len(tree))
            except StopIteration:
                minimum = middle + 1
        logger.info("The smallest specification is of size %s.", len(tree))
        return self._get_specification_rules(label, tree)


class RuleDB(RuleDBBase):
    def __init__(self) -> None:
        super().__init__()
        self._rule_to_strategy: Dict[Tuple[int, Tuple[int, ...]], AbstractStrategy] = {}

    @property
    def rule_to_strategy(self) -> Dict[Tuple[int, Tuple[int, ...]], AbstractStrategy]:
        return self._rule_to_strategy



"""A class for automatically performing combinatorial exploration."""
import gc
import logging
import platform
import time
import warnings
from collections import defaultdict
from datetime import timedelta
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import logzero
import tabulate
from logzero import logger
from sympy import Eq, Function, var

from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CSSstrategy,
    SpecificationClassesAndStrats,
    WorkPacket,
)

from .class_db import ClassDB
from .class_queue import DefaultQueue
from .exception import (
    ExceededMaxtimeError,
    InvalidOperationError,
    SpecificationNotFound,
    StrategyDoesNotApply,
)
from .rule_db import RuleDB, RuleDBForgetStrategy
from .rule_db.base import RuleDBBase
from .specification import CombinatorialSpecification
from .strategies import (
    AbstractStrategy,
    StrategyFactory,
    StrategyPack,
    VerificationRule,
)
from .strategies.rule import AbstractRule
from .utils import (
    cssiteratortimer,
    cssmethodtimer,
    get_mem,
    nice_pypy_mem,
    size_to_readable,
)

if platform.python_implementation() == "CPython":
    from pympler.asizeof import asizeof

__all__ = ["CombinatorialSpecificationSearcher"]

warnings.simplefilter("once", Warning)

logzero.loglevel(logging.INFO)


class CombinatorialSpecificationSearcher(Generic[CombinatorialClassType]):
    """
    The CombinatorialSpecificationSearcher class.

    This is used to build up knowledge about a combinatorial_class with respect
    to the given strategies and search for a combinatorial specification.
    """

    def __init__(
        self,
        start_class: CombinatorialClassType,
        strategy_pack: StrategyPack,
        ruledb: Optional[Union[str, RuleDB]] = None,
        **kwargs,
    ):
        """
        Initialise CombinatorialSpecificationSearcher.

        OTHER INPUT:
            - `ruledb`: a string to specify the type of ruledb to use for the
            search. Default to `None` but can be changed to "forget" for a ruledb that
            saves more memory.
            - `expand_verified`: if True, every verified combinatorial class will
              still be expanded using the strategies in strategy pack
            - `debug`: if True every rule found will be sanity checked and logged
              to logging.DEBUG
            - `function_kwargs` are passed to the call method of strategies
        """
        self.strategy_pack = strategy_pack
        self.debug = kwargs.get("debug", False)
        self.expand_verified = kwargs.get("expand_verified", False)
        if self.debug:
            logzero.loglevel(logging.DEBUG, True)
        self.kwargs = kwargs.get("function_kwargs", dict())

        self.func_times: Dict[str, float] = defaultdict(float)
        self.func_calls: Dict[str, int] = defaultdict(int)

        self.kwargs["symmetry"] = bool(strategy_pack.symmetries)

        self.classdb = ClassDB[CombinatorialClassType](type(start_class))
        self.classqueue = DefaultQueue(strategy_pack)

        if ruledb is None:
            self.ruledb: RuleDBBase = RuleDB()
        elif ruledb == "forget":
            self.ruledb = RuleDBForgetStrategy(self.classdb, self.strategy_pack)
        elif isinstance(ruledb, RuleDBBase):
            self.ruledb = ruledb
        else:
            raise ValueError(
                "ruledb argument should be None or 'forget' or a RuleDB object"
            )

        # initialise the run with start_class
        self.start_label = self.classdb.get_label(start_class)
        self._add_to_queue(self.start_label)
        self.tried_to_verify: Set[int] = set()
        self.symmetry_expanded: Set[int] = set()
        self.try_verify(start_class, self.start_label)
        if self.symmetries:
            self._symmetry_expand(start_class, self.start_label)

    @property
    def verification_strategies(self) -> Sequence[CSSstrategy]:
        """The verification strategies from the strategy pack."""
        return self.strategy_pack.ver_strats

    @property
    def iterative(self) -> bool:
        """The iterative parameter from the strategy pack."""
        return self.strategy_pack.iterative

    @property
    def symmetries(self) -> Sequence[CSSstrategy]:
        """The symmetries functions for the strategy pack."""
        return self.strategy_pack.symmetries

    def try_verify(self, comb_class: CombinatorialClassType, label: int) -> None:
        """
        Try to verify the combinatorial class.
        """
        if label not in self.tried_to_verify:
            for strategy in self.verification_strategies:
                if self.ruledb.is_verified(label):
                    break
                for start_label, end_labels, rule in self._expand_class_with_strategy(
                    comb_class, strategy, label
                ):
                    self._add_rule(start_label, end_labels, rule)
            self.tried_to_verify.add(label)

    @cssmethodtimer("is empty")
    def is_empty(self, comb_class: CombinatorialClassType, label: int) -> bool:
        """Return True if a combinatorial class contains no objects, False
        otherwise."""
        empty = self.classdb.is_empty(comb_class, label)
        return empty

    def _expand(
        self,
        comb_class: CombinatorialClassType,
        label: int,
        strategies: Tuple[CSSstrategy, ...],
        inferral: bool,
    ) -> None:
        """
        Will expand the combinatorial class with given label using the given
        strategies.
        """
        if inferral:
            self._inferral_expand(comb_class, label, strategies)
        else:
            for strategy_generator in strategies:
                for start_label, end_labels, rule in self._expand_class_with_strategy(
                    comb_class, strategy_generator, label
                ):
                    self._add_rule(start_label, end_labels, rule)

    def _rules_from_strategy(
        self, comb_class: CombinatorialClassType, strategy: CSSstrategy
    ) -> Iterator[AbstractRule]:
        """Yield all the rules given by a strategy/strategy generator."""
        if isinstance(strategy, AbstractStrategy):
            try:
                yield strategy(comb_class, **self.kwargs)
            except StrategyDoesNotApply:
                pass
        elif isinstance(strategy, StrategyFactory):
            for strat in strategy(comb_class, **self.kwargs):
                if isinstance(strat, AbstractRule):
                    yield strat
                elif isinstance(strat, AbstractStrategy):
                    try:
                        yield strat(comb_class, **self.kwargs)
                    except StrategyDoesNotApply:
                        continue
                else:
                    raise InvalidOperationError(
                        "Attempting to add non Rule type. A Strategy "
                        "Generator's __call__ method should yield Strategy or "
                        "Strategy(comb_class, children) object."
                    )
        else:
            raise InvalidOperationError(
                "CSS can only expand a combinatorial class with "
                "Strategy and StrategyFactory"
            )

    @cssiteratortimer("_expand_class_with_strategy")
    def _expand_class_with_strategy(
        self,
        comb_class: CombinatorialClassType,
        strategy_generator: CSSstrategy,
        label: Optional[int] = None,
        initial: bool = False,
    ) -> Iterator[Tuple[int, Tuple[int, ...], AbstractRule]]:
        """
        Will expand the class with given strategy. Return time taken.
        """
        logger.debug(
            "Expanding label %s with %s",
            label,
            strategy_generator,
        )
        if label is None:
            label = self.classdb.get_label(comb_class)

        for rule in self._rules_from_strategy(comb_class, strategy_generator):
            try:
                children = rule.children
            except StrategyDoesNotApply:
                continue
            if len(children) == 1 and rule.comb_class == children[0]:
                logger.debug(
                    "The equivalence strategy %s returned the same "
                    "combinatorial class when applied to %r",
                    str(rule).split(" ")[1],
                    comb_class,
                )
                continue
            end_labels = [self.classdb.get_label(child) for child in children]
            if rule.comb_class == comb_class:
                start_label = label
            else:
                start_label = self.classdb.get_label(rule.comb_class)

            # TODO: observe that creating this constructor could be costly,
            # e.g. Cartesian
            if self.debug:
                logger.debug(
                    "Adding combinatorial rule %s -> %s\n%s",
                    start_label,
                    tuple(end_labels),
                    rule,
                )
                try:
                    n = 4
                    for i in range(n + 1):
                        rule.sanity_check(n=i)
                    logger.debug("Sanity checked rule to length %s.", n)
                except NotImplementedError as e:
                    logger.debug(
                        "Could not sanity check rule due to:\n"
                        "NotImplementedError: %s",
                        e,
                    )
            if any(
                self.ruledb.are_equivalent(start_label, elabel) for elabel in end_labels
            ):
                # This says comb_class = comb_class, so we skip it, but mark
                # every other class as empty.
                for elabel in end_labels:
                    if not self.ruledb.are_equivalent(start_label, elabel):
                        self._add_empty_rule(elabel)
                if self.debug:
                    for elabel, child in zip(end_labels, children):
                        if not self.ruledb.are_equivalent(start_label, elabel):
                            assert child.is_empty()

            yield start_label, tuple(end_labels), rule

    def _add_rule(
        self, start_label: int, end_labels: Tuple[int, ...], rule: AbstractRule
    ) -> None:
        """
        Add the cleaned rules labels.

        - try to verify children combinatorial classes
        - set workability of combinatorial classes
        - remove empty combinatorial classes
        - symmetry expand combinatorial classes
        - add class to classqueue
        """
        cleaned_end_labels = []
        for comb_class, child_label in zip(rule.children, end_labels):
            if self.symmetries and child_label not in self.symmetry_expanded:
                self._symmetry_expand(
                    comb_class, child_label
                )  # TODO: mark symmetries as empty where appropriate
            # Only applying is_empty check to comb classes that are
            # possibly empty.
            if rule.possibly_empty and self.is_empty(comb_class, child_label):
                logger.debug("Label %s is empty.", child_label)
                continue
            if rule.workable:
                self._add_to_queue(child_label)
            if not rule.inferrable:
                self._not_inferrable(child_label)
            if not rule.possibly_empty:
                self.classdb.set_empty(child_label, empty=False)

            self.try_verify(comb_class, child_label)

            cleaned_end_labels.append(child_label)

        if cleaned_end_labels == [start_label]:
            return

        if rule.ignore_parent:
            self._stop_yielding(start_label)

        if not cleaned_end_labels:
            # this must be a verification strategy!
            assert isinstance(rule, VerificationRule), rule.formal_step
        self.ruledb.add(start_label, tuple(cleaned_end_labels), rule)

    def _add_empty_rule(self, label: int) -> None:
        """Mark label as empty. Treated as verified as can count empty set."""
        self.classdb.set_empty(label, empty=True)

    def _symmetry_expand(self, comb_class: CombinatorialClassType, label: int) -> None:
        """Add symmetries of combinatorial class to the database."""
        sym_labels = set([label])
        for strategy_generator in self.symmetries:
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy_generator, label=label
            ):
                sym_label = end_labels[0]
                self.ruledb.add(start_label, (sym_label,), rule)
                self._stop_yielding(sym_label)
                sym_labels.add(sym_label)
        self.symmetry_expanded.update(sym_labels)

    def _inferral_expand(
        self,
        comb_class: CombinatorialClassType,
        label: int,
        inferral_strategies: Tuple[CSSstrategy, ...],
        skip: Optional[CSSstrategy] = None,
    ):
        """
        Inferral expand combinatorial class with given label and inferral
        strategies.

        It will apply all inferral strategies to an combinatorial class.
        Return True if combinatorial class is inferred.
        """
        if self.debug:
            assert comb_class == self.classdb.get_class(label)
        for i, strategy_generator in enumerate(inferral_strategies):
            if strategy_generator == skip:
                continue
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy_generator, label=label
            ):
                inf_class = rule.children[0]
                inf_label = end_labels[0]
                self._add_rule(start_label, end_labels, rule)
                self._not_inferrable(start_label)
                inferral_strategies = (
                    inferral_strategies[i + 1 :] + inferral_strategies[0 : i + 1]
                )
                self._inferral_expand(
                    inf_class, inf_label, inferral_strategies, skip=strategy_generator
                )
                break
        self._not_inferrable(label)

    def get_equations(self, **kwargs) -> Set[Eq]:
        """
        Returns a set of equations for all rules currently found.
        """
        x = var("x")

        def get_function(comb_class):
            label = self.classdb.get_label(comb_class)
            eqv_label = self.ruledb.equivdb[label]
            return Function("F_{}".format(eqv_label))(x)

        eqs = set()
        for start, ends, strategy in self.ruledb.all_rules():
            parent = self.classdb.get_class(start)
            children = tuple(map(self.classdb.get_class, ends))
            rule = strategy(parent, children)
            try:
                eq = rule.get_equation(get_function)
            except NotImplementedError:
                logger.info(
                    "can't find generating function for %s." " The comb class is:\n%s",
                    get_function(rule.comb_class),
                    rule.comb_class,
                )
                eq = Eq(get_function(rule.comb_class), Function("NOTIMPLEMENTED")(x))
            eqs.add(eq)
        return eqs

    def do_level(self) -> None:
        """Expand combinatorial classes in current queue. Combintorial classes
        found added to next."""
        for label, strategies, inferral in self.classqueue.do_level():
            comb_class = self.classdb.get_class(label)
            self._expand(comb_class, label, strategies, inferral)

    def _labels_to_expand(self) -> Iterator[WorkPacket]:
        yield from self.classqueue

    def _add_to_queue(self, label: int):
        self.classqueue.add(label)

    def _not_inferrable(self, label: int):
        self.classqueue.set_not_inferrable(label)

    def _stop_yielding(self, label: int):
        self.classqueue.set_stop_yielding(label)

    @cssmethodtimer("status")
    def status(self, elaborate: bool) -> str:
        """
        Return a string of the current status of the CombSpecSearcher.

        It includes:
        - number of combinatorial classes, and information about verification
        - the times spent in each of the main functions

        "elaborate" status updates are those that provide information that
        may be slow to compute
        """
        status = "CSS status:\n"
        total = sum(self.func_times.values())
        status += f"\tTotal time accounted for: {timedelta(seconds=int(total))}\n"
        status += self._css_status(total)
        status += self.classdb.status() + "\n"
        status += self.classqueue.status() + "\n"
        status += self.ruledb.status(elaborate) + "\n"
        status += self.mem_status(elaborate)
        return status

    @cssmethodtimer("status")
    def _css_status(self, total: float) -> str:
        table: List[Tuple[str, str, timedelta, str]] = []
        for explanation in self.func_calls:
            count = f"{self.func_calls[explanation]:,d}"
            time_spent = timedelta(seconds=int(self.func_times[explanation]))
            if total == 0:
                percentage = "? %"
            else:
                percentage = f"{int((self.func_times[explanation] * 100) / total)}%"
            table.append((explanation, count, time_spent, percentage))
        table.sort(key=lambda row: row[2], reverse=True)
        headers = ["", "Number of \napplications", "\nTime spent", "\nPercentage"]
        colalign = ("left", "right", "right", "right")
        return (
            "    "
            + tabulate.tabulate(table, headers=headers, colalign=colalign).replace(
                "\n", "\n    "
            )
            + "\n"
        )

    @cssmethodtimer("status")
    def mem_status(self, elaborate: bool) -> str:
        """Provide status information related to memory usage."""

        status = "Memory Status:\n"
        table: List[Tuple[str, str]] = []
        table.append(("OS Allocated", size_to_readable(get_mem())))
        if platform.python_implementation() == "CPython":
            # Warning: "asizeof" can be very slow!
            if elaborate:
                table.append(("CSS", size_to_readable(asizeof(self))))
                table.append(("ClassDB", size_to_readable(asizeof(self.classdb))))
                table.append(("ClassQueue", size_to_readable(asizeof(self.classqueue))))
                table.append(("RuleDB", size_to_readable(asizeof(self.ruledb))))

        elif platform.python_implementation() == "PyPy":
            gc_stats = cast(Any, gc.get_stats())
            stats = [
                ("Current Memory Used", gc_stats.total_gc_memory),
                ("Current Memory Allocated", gc_stats.total_allocated_memory),
                ("Current JIT Memory Used", gc_stats.jit_backend_used),
                ("Current JIT Memory Allocated", gc_stats.jit_backend_allocated),
                ("Peak Memory Used", gc_stats.peak_memory),
                ("Peak Memory Allocated Memory Used", gc_stats.peak_allocated_memory),
            ]
            for (desc, mem) in stats:
                table.append((desc, nice_pypy_mem(mem)))
        status += "    "
        status += tabulate.tabulate(table, colalign=("left", "right")).replace(
            "\n", "\n    "
        )
        status += "\n"
        if platform.python_implementation() == "PyPy":
            status += "\tTotal Garbage Collection Time: {}\n".format(
                timedelta(seconds=int(gc_stats.total_gc_time / 1000))
            )
        return status

    def run_information(self) -> str:
        """Return string detailing what CombSpecSearcher is looking for."""
        start_string = (
            "Initialising CombSpecSearcher for the combinatorial"
            " class:\n{}\n".format(self.classdb.get_class(self.start_label))
        )
        start_string += str(self.strategy_pack)
        return start_string

    def _log_spec_found(
        self, specification: CombinatorialSpecification, start_time: float
    ):
        found_string = "Specification built {}\n".format(
            time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime())
        )
        time_taken = time.time() - start_time
        found_string += f"Time taken: {timedelta(seconds=int(time_taken))}\n"
        found_string += self.status(elaborate=True)
        found_string += (
            f"Specification found has {specification.number_of_rules()} rules"
        )
        logger.info(found_string)

    def _log_status(self, start_time: float, status_update: int) -> None:
        time_taken = time.time() - start_time
        status = f"Time taken so far: {timedelta(seconds=int(time_taken))}\n"
        elaborate = time.time() - start_time > 100 * self.func_times["status"]
        status_start = time.time()
        status += self.status(elaborate=elaborate)
        ne_goal = 100 * self.func_times["status"] - (time.time() - start_time)
        next_elaborate = round(ne_goal - (ne_goal % status_update) + status_update)
        if elaborate:
            status += " -- status update took {} seconds --\n".format(
                round(time.time() - status_start, 2)
            )
        else:
            status += (
                " -- next elaborate status update in "
                f"{timedelta(seconds=next_elaborate)} --\n"
            )
        logger.info(status)

    def auto_search(self, **kwargs) -> CombinatorialSpecification:
        """
        An automatic search function.

        Classes will be expanded until a proof tree is found. A tree will be
        searched for approximately 1% of the search time. This can be set using
        the 'perc' keyword, as some percentage between 0 and 100.

        The search will continue, unless a proof tree is found. You can set the
        keyword 'max_time' to stop the search after 'max_time' many seconds.
        If max_time is reached it will raise a ExceededMaxtimeError.

        Information is logged to logger.info. It will also log the proof tree,
        in json format. For periodic status_updates, set the keyword flag
        'status_update', an update will be given every status_update seconds.
        "Elaborate" status updates, which provide slow-to-calculate information
        will be given periodically, so that they take less than 1% of computation
        time.

        If 'smallest' is set to 'True' then the searcher will return a proof
        tree that is as small as possible.
        """
        auto_search_start = time.time()

        perc = kwargs.get("perc", 1)
        if not 0 < perc <= 100:
            logger.warning(
                (
                    "Percentage not between 0 and 100, so assuming 1%"
                    " search percentage."
                ),
            )
            perc = 1
        status_update = kwargs.get("status_update", None)
        max_time = kwargs.get("max_time", None)
        status_start = time.time()
        start_string = "Auto search started {}\n".format(
            time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime())
        )
        start_string += self.run_information()
        logger.info(start_string)

        max_expansion_time = 0
        expanding = True

        while expanding:
            expanding, status_start = self._expand_classes_for(
                max_expansion_time, status_update, status_start, auto_search_start
            )

            spec_search_start = time.time()
            logger.debug("Searching for specification.")
            specification = self.get_specification(
                smallest=kwargs.get("smallest", False),
                minimization_time_limit=0.01 * (time.time() - auto_search_start),
            )
            if specification is not None:
                self._log_spec_found(specification, auto_search_start)
                return specification
            logger.debug("No specification found.")
            if max_time is not None:
                if time.time() - auto_search_start > max_time:
                    raise ExceededMaxtimeError(
                        "Exceeded maximum time. Aborting auto search.",
                    )
            # worst case, search every hour
            multiplier = 100 // perc
            max_expansion_time = min(
                multiplier * (time.time() - spec_search_start), 3600
            )
            logger.debug(
                "Will expand for %s seconds.",
                round(max_expansion_time, 2),
            )

    def _auto_search_rules(self) -> SpecificationClassesAndStrats:
        """
        A basic auto search for returning equivalence paths and rules.

        This method is used by CombinatorialSpecification for expanding
        verified classes.

        Will raise SpecificationNotFound error if no specification is found
        after running out of classes to expand.
        """
        status_start = time.time()
        status_update = None  # this prevents status updates happening
        auto_search_start = time.time()
        max_expansion_time: float = 0
        expanding = True
        while expanding:
            expanding, status_start = self._expand_classes_for(
                max_expansion_time, status_update, status_start, auto_search_start
            )
            spec_search_start = time.time()
            spec = self._get_specification_rules(
                0.01 * (time.time() - auto_search_start)
            )
            if spec is not None:
                return spec
            max_expansion_time = min(0.01 * (time.time() - spec_search_start), 3600)
        raise SpecificationNotFound

    def _expand_classes_for(
        self,
        expansion_time: float,
        status_update: Optional[int],
        status_start: float,
        auto_search_start: float,
    ) -> Tuple[bool, float]:
        """
        Will expand classes for `expansion_time` seconds.

        It will return a pair (bool, time), where the bool is True if there are
        more classes to expand, False otherwise. The `time` is the initial time
        for checking whether to post a status update.
        """
        expansion_start = time.time()
        last_label = None
        expanding = True
        for label, strategies, inferral in self._labels_to_expand():
            if label != last_label:
                comb_class = self.classdb.get_class(label)
                last_label = label
            if self.expand_verified or not self.ruledb.is_verified(label):
                self._expand(comb_class, label, strategies, inferral)
            if time.time() - expansion_start > expansion_time:
                break
            if status_update is not None and time.time() - status_start > status_update:
                self._log_status(auto_search_start, status_update)
                status_start = time.time()
        else:
            expanding = False
            logger.info("No more classes to expand.")
        return expanding, status_start

    @cssmethodtimer("get specification")
    def get_specification(
        self, minimization_time_limit: float = 10, smallest: bool = False
    ) -> Optional[CombinatorialSpecification]:
        """
        Return a CombinatorialSpecification if the universe contains one.

        The minimization_time_limit only applies when smallest is false.

        The function will return None if no such CombinatorialSpecification
        exists in the universe.
        """
        spec = self._get_specification_rules(minimization_time_limit, smallest)
        if spec is None:
            return None
        start_class = self.classdb.get_class(self.start_label)
        strategies, comb_class_eqv_paths = spec
        logger.info("Creating a specification.")
        return CombinatorialSpecification(start_class, strategies, comb_class_eqv_paths)

    @cssmethodtimer("get specification")
    def _get_specification_rules(
        self, minimization_time_limit: float = 10, smallest: bool = False
    ) -> Optional[SpecificationClassesAndStrats]:
        """
        Returns the equivalence paths needed to create a
        CombinatorialSpecification, if one exists.

        The minimization_time_limit only applies when smallest is false.

        The function will return None if no such CombinatorialSpecification
        exists in the universe.
        """
        try:
            if smallest:
                if self.iterative:
                    raise InvalidOperationError("can't use iterative and smallest")
                rules, eqv_paths = self.ruledb.get_smallest_specification(
                    self.start_label
                )
            else:
                rules, eqv_paths = self.ruledb.get_specification_rules(
                    self.start_label,
                    minimization_time_limit=minimization_time_limit,
                    iterative=self.iterative,
                )
        except SpecificationNotFound:
            return None
        comb_class_eqv_paths = [
            list(map(self.classdb.get_class, path)) for path in eqv_paths
        ]
        strategies = [(self.classdb.get_class(label), rule) for label, rule in rules]
        return strategies, comb_class_eqv_paths



#!/usr/bin/env python
import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="comb_spec_searcher",
    version="2.4.0",
    author="Permuta Triangle",
    author_email="permutatriangle@gmail.com",
    description="A library for performing combinatorial exploration.",
    license="BSD-3",
    keywords="enumerative combinatorics combinatorial specification counting",
    url="https://github.com/PermutaTriangle/comb_spec_searcher",
    project_urls={
        "Source": "https://github.com/PermutaTriangle/comb_spec_searcher",
        "Tracker": ("https://github.com/PermutaTriangle/comb_spec_searcher" "/issues"),
    },
    packages=find_packages(),
    package_data={"comb_spec_searcher": ["py.typed"]},
    long_description=read("README.rst"),
    python_requires=">=3.6",
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    install_requires=[
        "logzero==1.6.2",
        "sympy==1.6.2",
        "psutil==5.7.3",
        "pympler==0.9",
        "requests==2.24.0",
        "typing-extensions==3.7.4.3",
        "tabulate==0.8.7",
    ],
)



"""
The rule class is used for a specific application of a strategy on a combinatorial
class. This is not something the user should implement, as it is just a wrapper for
calling the Strategy class and storing its results.

A CombinatorialSpecification is (more or less) a set of Rule.
"""
import abc
from collections import defaultdict
from itertools import chain, product
from random import randint
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    cast,
)

from sympy import Eq, Function

from comb_spec_searcher.typing import (
    Objects,
    ObjectsCache,
    SubObjects,
    SubTerms,
    Terms,
    TermsCache,
)

from ..combinatorial_class import CombinatorialClassType, CombinatorialObjectType
from ..exception import SanityCheckFailure, StrategyDoesNotApply
from .constructor import Constructor, DisjointUnion

if TYPE_CHECKING:
    from .strategy import AbstractStrategy, Strategy, VerificationStrategy
    from .strategy_pack import StrategyPack

__all__ = ("Rule", "VerificationRule")


class AbstractRule(abc.ABC, Generic[CombinatorialClassType, CombinatorialObjectType]):
    """
    An instance of Rule is created by the __call__ method of strategy.
    """

    def __init__(
        self,
        strategy: "AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        self.comb_class = comb_class
        self._strategy = strategy
        self.terms_cache: TermsCache = []
        self.objects_cache: ObjectsCache = []
        self.subrecs: Optional[Tuple[Callable[..., int], ...]] = None
        self.subgenerators: Optional[
            Tuple[Callable[..., Iterator[CombinatorialObjectType]], ...]
        ] = None
        self.subsamplers: Optional[Tuple[Callable[..., CombinatorialObjectType], ...]]
        self.subterms: Optional[SubTerms] = None
        self.subobjects: Optional[SubObjects] = None
        self._children = children
        self._non_empty_children: Optional[Tuple[CombinatorialClassType, ...]] = None

    @property
    def strategy(
        self,
    ) -> "AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]":
        return self._strategy

    @property
    def ignore_parent(self) -> bool:
        """
        Return True if it is not worth expanding the parent/comb_class if
        this rule is found.
        """
        return self._strategy.ignore_parent

    @property
    def inferrable(self) -> bool:
        """
        Return True if the children could change using inferral strategies.
        """
        return self._strategy.inferrable

    @property
    def possibly_empty(self) -> bool:
        """
        Return True if it is possible that a child is empty.
        """
        return self._strategy.possibly_empty

    @property
    def workable(self) -> bool:
        """
        Return True if the children can expanded using other strategies.
        """
        return self._strategy.workable

    def set_subrecs(
        self,
        get_subrule: Callable[
            [CombinatorialClassType],
            "AbstractRule[CombinatorialClassType, CombinatorialObjectType]",
        ],
    ) -> None:
        """
        In general a rule comes in the context of a set of rules, e.g. in a
        CombinatorialSpecification. In order to count, generate, etc using the
        rule we must set the subfunctions that the children refer to.
        """
        self.subrecs = tuple(
            get_subrule(child).count_objects_of_size for child in self.children
        )
        self.subsamplers = tuple(
            get_subrule(child).random_sample_object_of_size for child in self.children
        )
        self.subobjects = tuple(
            get_subrule(child).get_objects for child in self.children
        )
        self.subterms = tuple(get_subrule(child).get_terms for child in self.children)

    @property
    def children(self) -> Tuple[CombinatorialClassType, ...]:
        """
        Return the tuple of CombinatorialClass that are found by applying the
        decomposition function of the strategy to the parent.
        """
        if self._children is None:
            self._children = self._strategy.decomposition_function(self.comb_class)
            if self._children is None:
                raise StrategyDoesNotApply("{} does not apply".format(self._strategy))
        return self._children

    @property
    def formal_step(self) -> str:
        """
        Return a string with a short explanation about the rule.
        """
        return self._strategy.formal_step()

    @abc.abstractmethod
    def is_equivalence(self) -> bool:
        """
        Returns True if the rule is an equivalence.
        """

    def non_empty_children(self) -> Tuple[CombinatorialClassType, ...]:
        """
        Return the tuple of non-empty combinatorial classes that are found
        by applying the decomposition function.
        """
        if self._non_empty_children is None:
            self._non_empty_children = tuple(
                child for child in self.children if not child.is_empty()
            )
        return self._non_empty_children

    @abc.abstractmethod
    def _ensure_level(self, n: int) -> None:
        """
        Ensures the terms are computed and in the terms_cache upto size n.
        """

    def get_terms(self, n: int) -> Terms:
        """
        Return the terms for the given n.
        """
        self._ensure_level(n)
        return self.terms_cache[n]

    def count_objects_of_size(self, n: int, **parameters: int) -> int:
        """
        The function count the objects with respect to the parameters. The
        result is cached.
        """
        terms = self.get_terms(n)
        params_tuple = tuple(parameters[k] for k in self.comb_class.extra_parameters)
        return terms[params_tuple]

    @abc.abstractmethod
    def get_equation(
        self, get_function: Callable[[CombinatorialClassType], Function]
    ) -> Eq:
        """
        Return the equation for the (ordinary) generating function.
        """

    @abc.abstractmethod
    def _ensure_level_objects(self, n: int) -> None:
        """
        Ensures the objects are computed and in the objects cache upto size n.
        """

    def get_objects(self, n: int) -> Objects:
        """
        Return the objects for the given n.
        """
        self._ensure_level_objects(n)
        return self.objects_cache[n]

    def generate_objects_of_size(
        self, n: int, **parameters: int
    ) -> Iterator[CombinatorialObjectType]:
        """
        Generate the objects by using the underlying bijection between the
        parent and children.
        """
        objects = self.get_objects(n)
        params_tuple = tuple(parameters[k] for k in self.comb_class.extra_parameters)
        yield from objects[params_tuple]

    @abc.abstractmethod
    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        """Return a random objects of the give size."""

    @abc.abstractmethod
    def sanity_check(self, n: int) -> bool:
        """
        Sanity check that this is a valid rule.

        Raise a SanityCheckFailure error if the sanity_check fails.
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractRule):
            return NotImplemented
        return (
            isinstance(other, self.__class__)
            and self.comb_class == other.comb_class
            and self._strategy == other._strategy
        )

    def __str__(self) -> str:
        def frontpad(res, height):
            n = max(len(s) for s in res)
            for _ in range(height - len(res)):
                res.append(" " * n)

        def backpad(res):
            n = max(len(s) for s in res)
            for i, s in enumerate(res):
                res[i] = s + " " * (n - len(s))

        def join(res, new):
            backpad(new)
            backpad(res)
            frontpad(res, len(new))
            for i, s in enumerate(new):
                res[i] += s

        res = str(self.comb_class).split("\n")
        backpad(res)
        if isinstance(self, Rule):
            children = [str(child).split("\n") for child in self.children]
            symbol_height = min(1, len(res) - 1)
            eq_symbol = (
                ["     " for i in range(symbol_height)]
                + ["  {}  ".format(self.strategy.get_eq_symbol())]
                + ["     " for i in range(symbol_height)]
            )
            join(res, eq_symbol)
            join(res, children[0])
            if len(children) > 1:
                op_symbol = (
                    ["     " for i in range(symbol_height)]
                    + ["  {}  ".format(self.strategy.get_op_symbol())]
                    + ["     " for i in range(symbol_height)]
                )
                for child in children[1:]:
                    join(res, op_symbol)
                    join(res, child)
        return f"{self.formal_step}\n" + "\n".join(x for x in res)


class Rule(AbstractRule[CombinatorialClassType, CombinatorialObjectType]):
    def __init__(
        self,
        strategy: "Strategy[CombinatorialClassType, CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        super().__init__(strategy, comb_class, children)
        self._constructor: Optional[Constructor] = None

    @property
    def strategy(self) -> "Strategy[CombinatorialClassType, CombinatorialObjectType]":
        return cast(
            "Strategy[CombinatorialClassType, CombinatorialObjectType]",
            super().strategy,
        )

    @property
    def constructor(
        self,
    ) -> Constructor[CombinatorialClassType, CombinatorialObjectType]:
        """
        Return the constructor, that contains all the information about how to
        count/generate objects from the rule.
        """
        if self._constructor is None:
            self._constructor = self.strategy.constructor(
                self.comb_class, self.children
            )
            if self._constructor is None:
                raise StrategyDoesNotApply("{} does not apply".format(self.strategy))
        return self._constructor

    def is_equivalence(self):
        return self.strategy.can_be_equivalent() and len(self.non_empty_children()) == 1

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        """
        This encodes the backward map of the underlying bijection that the
        strategy implies.
        """
        yield from self.strategy.backward_map(self.comb_class, objs, self.children)

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        """
        This encodes the forward map of the underlying bijection that the
        strategy implies.
        """
        return self.strategy.forward_map(self.comb_class, obj, self.children)

    def to_equivalence_rule(self) -> "EquivalenceRule":
        """
        Return the reverse rule. At this stage, reverse rules can only be
        created for equivalence rules.
        """
        assert (
            self.is_equivalence()
        ), f"EquivalenceRule can only be created for equivalence rules\n{self}"
        return EquivalenceRule(self)

    def to_reverse_rule(self) -> "ReverseRule":
        """
        Return the reverse rule. At this stage, reverse rules can only be
        created for equivalence rules.
        """
        assert (
            self.is_equivalence()
        ), "reverse rule can only be created for equivalence rules"
        return ReverseRule(self)

    def _ensure_level(self, n: int) -> None:
        if self.subterms is None:
            raise RuntimeError("set_subrecs must be set first")
        while n >= len(self.terms_cache):
            terms = self.constructor.get_terms(self.subterms, len(self.terms_cache))
            self.terms_cache.append(terms)

    def get_equation(
        self, get_function: Callable[[CombinatorialClassType], Function]
    ) -> Eq:
        """
        Return the equation for the (ordinary) generating function.
        """
        lhs_func = get_function(self.comb_class)
        rhs_funcs = tuple(get_function(comb_class) for comb_class in self.children)
        return self.constructor.get_equation(lhs_func, rhs_funcs)

    def _ensure_level_objects(self, n: int) -> None:
        if self.subobjects is None:
            raise RuntimeError("set_subrecs must be set first")
        while n >= len(self.objects_cache):
            objects: Objects = defaultdict(list)
            for parameters, subobjects in self.constructor.get_sub_objects(
                self.subobjects, len(self.objects_cache)
            ):
                for sub_objs in product(*subobjects):
                    objects[parameters].extend(self.backward_map(sub_objs))
            self.objects_cache.append(objects)

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        assert self.subrecs is not None, "you must call the set_subrecs function first"
        assert (
            self.subsamplers is not None
        ), "you must call the set_subrecs function first"
        total_count = self.count_objects_of_size(n=n, **parameters)
        subobjs = self.constructor.random_sample_sub_objects(
            total_count, self.subsamplers, self.subrecs, n, **parameters
        )
        objs = tuple(self.backward_map(subobjs))
        idx = randint(0, len(objs) - 1)
        return objs[idx]

    def sanity_check(self, n: int) -> bool:
        try:
            return self._sanity_check_count(n) and self._sanity_check_objects(n)
        except SanityCheckFailure as e:
            raise e

    def _sanity_check_count(self, n: int) -> bool:
        """
        Sanity check that the count given by the rule matches the brute force count.
        """
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        actual_terms = self.comb_class.get_terms(n)
        temp_subterms = self.subterms
        self.subterms = tuple(child.get_terms for child in self.children)
        rule_terms = self.get_terms(n)
        self.subterms = temp_subterms
        if actual_terms != rule_terms:
            raise SanityCheckFailure(
                f"The following rule failed sanity check:\n"
                f"{self}\n"
                f"Failed for size {n}\n"
                f"The actual count is {actual_terms}.\n"
                f"The rule count is {rule_terms}.",
            )
        return True

    def _sanity_check_objects(self, n: int) -> bool:
        """
        Sanity check that the object given by the rule matches the brute force
        generated objects.
        """
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        tempobjects = self.subobjects
        self.subobjects = tuple(child.get_objects for child in self.children)
        try:
            rule_objects = self.get_objects(n)
        except NotImplementedError:
            # Skipping testing rules that have not implemented object generation.
            return True
        self.subobjects = tempobjects
        actual_objects = self.comb_class.get_objects(n)
        for obj_list in chain(rule_objects.values(), actual_objects.values()):
            obj_list.sort()
        if actual_objects != rule_objects:
            raise SanityCheckFailure(
                f"The following rule failed sanity check:\n"
                f"{self}\n"
                f"Failed for size {n}:\n"
                f"The actual objects are {actual_objects}\n"
                f"The rule generated objects {rule_objects}."
            )
        return True


class EquivalenceRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    def __init__(self, rule: Rule):
        non_empty_children = rule.non_empty_children()
        assert rule.is_equivalence(), "not an equivalence rule: {}".format(str(rule))
        child = non_empty_children[0]
        super().__init__(rule.strategy, rule.comb_class, (child,))
        self.child_idx = rule.children.index(child)
        self.actual_children = rule.children
        self._constructor: Optional[DisjointUnion] = None

    @property
    def constructor(self) -> DisjointUnion:
        """
        Return the constructor, that contains all the information about how to
        count/generate objects from the rule.
        """
        if self._constructor is None:
            original_constructor = cast(
                DisjointUnion,
                self.strategy.constructor(self.comb_class, self.actual_children),
            )
            assert isinstance(original_constructor, DisjointUnion)
            self._constructor = DisjointUnion(
                self.comb_class,
                self.children,
                (original_constructor.extra_parameters[self.child_idx],),
            )
        return self._constructor

    @staticmethod
    def is_equivalence() -> bool:
        return True

    @property
    def formal_step(self) -> str:
        return "{} but only the child at index {} is non-empty".format(
            self.strategy.formal_step(), self.child_idx
        )

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        actual_objs = tuple(
            objs[0] if i == self.child_idx else None
            for i in range(len(self.actual_children))
        )
        yield from self.strategy.backward_map(
            self.comb_class, actual_objs, self.actual_children
        )

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        return (
            self.strategy.forward_map(self.comb_class, obj, self.actual_children)[
                self.child_idx
            ],
        )


class EquivalencePathRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    """
    A class for shortening a chain of equivalence rules into a single Rule.
    """

    def __init__(self, rules: Sequence[Rule]):
        assert all(rule.is_equivalence() for rule in rules)
        super().__init__(rules[0].strategy, rules[0].comb_class, rules[-1].children)
        self.rules = rules
        self._constructor: Optional[DisjointUnion] = None

    @property
    def constructor(self) -> DisjointUnion:
        if self._constructor is None:
            extra_parameters: Dict[str, str] = {
                k: k for k in self.comb_class.extra_parameters
            }
            for rule in self.rules:
                original_constructor = rule.constructor
                assert isinstance(original_constructor, DisjointUnion)
                rules_parameters = original_constructor.extra_parameters[0]
                extra_parameters = {
                    parent_var: rules_parameters[child_var]
                    for parent_var, child_var in extra_parameters.items()
                    if child_var in rules_parameters
                }
            fixed_values = {
                k: 0
                for k in self.children[0].extra_parameters
                if k not in extra_parameters.values()
            }
            self._constructor = DisjointUnion(
                self.comb_class, self.children, (extra_parameters,), (fixed_values,)
            )
        return self._constructor

    @staticmethod
    def is_equivalence() -> bool:
        return True

    @property
    def formal_step(self) -> str:
        return ", then ".join(rule.formal_step for rule in self.rules)

    def eqv_path_rules(self) -> List[Tuple[CombinatorialClassType, Rule]]:
        """
        Returns the list of (parent, rule) pairs that make up the equivalence
        path.
        """
        eqv_path_rules = []
        curr = self.comb_class
        for rule in self.rules:
            eqv_path_rules.append((curr, rule))
            curr = rule.children[0]
        return eqv_path_rules

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        res = cast(Tuple[CombinatorialObjectType], objs)
        for rule in reversed(self.rules):
            try:
                res = (next(rule.backward_map(res)),)
            except StopIteration:
                return
        yield res[0]

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        res = cast(CombinatorialObjectType, obj)
        for rule in reversed(self.rules):
            res = cast(CombinatorialObjectType, rule.forward_map(res)[0])
        return (res,)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and [r.strategy for r in self.rules] == [
            r.strategy for r in other.rules
        ]

    def __str__(self) -> str:
        def frontpad(res, height):
            n = max(len(s) for s in res)
            for _ in range(height - len(res)):
                res.append(" " * n)

        def backpad(res):
            n = max(len(s) for s in res)
            for i, s in enumerate(res):
                res[i] = s + " " * (n - len(s))

        def join(res, new):
            backpad(new)
            backpad(res)
            frontpad(res, len(new))
            for i, s in enumerate(new):
                res[i] += s

        res = str(self.comb_class).split("\n")
        backpad(res)
        comb_classes = [str(rule.children[0]).split("\n") for rule in self.rules]
        symbol_height = min(1, len(res) - 1)
        eq_symbol = (
            ["     " for i in range(symbol_height)]
            + ["  {}  ".format("=")]
            + ["     " for i in range(symbol_height)]
        )
        for comb_class in comb_classes:
            join(res, eq_symbol)
            join(res, comb_class)
        return "{}\n".format(self.formal_step) + "\n".join(x for x in res)


class ReverseRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    """
    A class for creating the reverse of an equivalence rule.
    """

    def __init__(self, rule: Rule):
        assert (
            rule.is_equivalence()
        ), "reversing a rule only works for equivalence rules"
        super().__init__(rule.strategy, rule.children[0], (rule.comb_class,))
        self.original_rule = rule
        self._constructor: Optional[DisjointUnion] = None

    @property
    def constructor(self) -> DisjointUnion:
        if self._constructor is None:
            original_constructor = self.original_rule.constructor
            assert isinstance(original_constructor, DisjointUnion), (
                "reverse rule coming from non disjoint union strategy - "
                "you'll need to update the ReverseRule constructor!"
            )
            flipped_extra_params = {
                b: a for a, b in original_constructor.extra_parameters[0].items()
            }
            fixed_values = {
                k: 0
                for k in self.children[0].extra_parameters
                if k not in flipped_extra_params.values()
            }
            self._constructor = DisjointUnion(
                self.comb_class, self.children, (flipped_extra_params,), (fixed_values,)
            )
        return self._constructor

    @staticmethod
    def is_equivalence() -> bool:
        return True

    @property
    def formal_step(self) -> str:
        return "reverse of '{}'".format(self.strategy.formal_step())

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        yield cast(CombinatorialObjectType, self.original_rule.forward_map(objs[0])[0])

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        return (next(self.original_rule.backward_map((obj,))),)


class VerificationRule(AbstractRule[CombinatorialClassType, CombinatorialObjectType]):
    """
    A class for verification rules. In particular, the children will be an
    empty tuple if it applies, else None.
    """

    def __init__(
        self,
        strat: "VerificationStrategy[CombinatorialClassType,CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        assert not children
        super().__init__(strat, comb_class, children)

    @property
    def strategy(
        self,
    ) -> "VerificationStrategy[CombinatorialClassType, CombinatorialObjectType]":
        return cast(
            "VerificationStrategy[CombinatorialClassType, CombinatorialObjectType]",
            super().strategy,
        )

    @staticmethod
    def is_equivalence() -> bool:
        return False

    def _ensure_level(self, n: int) -> None:
        while n >= len(self.terms_cache):
            terms = self.strategy.get_terms(self.comb_class, len(self.terms_cache))
            self.terms_cache.append(terms)

    def _ensure_level_objects(self, n: int) -> None:
        while n >= len(self.objects_cache):
            objects = self.strategy.get_objects(
                self.comb_class, len(self.objects_cache)
            )
            self.objects_cache.append(objects)

    def pack(self) -> "StrategyPack":
        return self.strategy.pack(self.comb_class)

    def get_equation(
        self,
        get_function: Callable[[CombinatorialClassType], Function],
        funcs: Optional[Dict[CombinatorialClassType, Function]] = None,
    ) -> Eq:
        lhs_func = get_function(self.comb_class)
        return Eq(lhs_func, self.strategy.get_genf(self.comb_class, funcs))

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        return self.strategy.random_sample_object_of_size(
            self.comb_class, n, **parameters
        )

    def sanity_check(self, n: int) -> bool:
        return self.get_terms(n) == self.comb_class.get_terms(n)
